#include "itkMultiThreader.h"
#include "itkSimpleFastMutexLock.h"
#include <stdlib.h>
#include <time.h>

template<class T>
class FindMax
{
public:
  /**
   * The given data is copied into internal storage.
   * @param data The data whose maximal value we are looking for.
   */
  void setData(std::vector<T> & data);

  /**
   * Find the maximal value in the given data. If the data was not set an exception
   * is thrown.
   * @param value The maximal value is returned in this parameter.
   * @return The index of the maximal value.
   */
  unsigned int find(T &value);
private:

  /**
   * This is the actual threaded function which does the searching for maximal
   * value.
   */
  static ITK_THREAD_RETURN_TYPE ThreaderCallback( void *arg );

  std::vector<T> data;
  unsigned int maxIndex;
  itk::SimpleFastMutexLock mutex;
};


int main(int argc, char *argv[])
{
  const unsigned int DATA_SIZE = 15;
  FindMax<unsigned int> fMax;
  unsigned int maxVal, maxIndex, i;
  
  srand((unsigned)time(NULL)); //seed random number generator   
  std::vector<unsigned int> myData;
  for(i=0; i<DATA_SIZE; i++)
     myData.push_back(rand());
   
   fMax.setData(myData);

   clock_t startTime = clock();

   maxIndex = fMax.find(maxVal);

   clock_t endTime = clock();
   std::cout<<"running time: "<<(double)(endTime-startTime)/CLOCKS_PER_SEC<<"s"<<std::endl;

   std::cout<<"data: "<<std::endl;
   for(i=0; i<myData.size(); i++)
     std::cout<<myData[i]<<" ";
   std::cout<<std::endl;

   std::cout<<"maximal value: "<<maxVal<<std::endl;
   std::cout<<"maximal value location: "<<maxIndex<<std::endl;
}


template<class T>
void FindMax<T>::setData(std::vector<T> & data)
{ 
  this->data.insert(this->data.begin(), data.begin(), data.end()); 
}


template<class T>
unsigned int FindMax<T>::find(T &value)
{
                   //create a new threader
  itk::MultiThreader::Pointer threader = itk::MultiThreader::New();

             //number of threads to use, number of cores - actually this is the default for the threader
             //it is only set here explicitly so that the user can change it.
  unsigned int numberOfThreads = itk::MultiThreader::GetGlobalDefaultNumberOfThreads();
  threader->SetNumberOfThreads(numberOfThreads);
            //initialize the search
  this->maxIndex = 0;
         //pass ourselves as the user info 
  threader->SetSingleMethod( FindMax<T>::ThreaderCallback, this);
            //run all threads and block till they finish
  threader->SingleMethodExecute();
  
  value = this->data[this->maxIndex];
  return this->maxIndex;
}


template<class T>
ITK_THREAD_RETURN_TYPE FindMax<T>::ThreaderCallback( void *arg )
{

  typedef itk::MultiThreader::ThreadInfoStruct ThreadInfoType;

  ThreadInfoType *infoStruct = static_cast<ThreadInfoType*>(arg);
  
                           //dynamic_cast doesn't work with void *
  FindMax<T> *caller = reinterpret_cast<FindMax<T> *>(infoStruct->UserData);
  if(caller!=NULL)
  {
    double threadID = static_cast<double>(infoStruct->ThreadID);
    double numThreads = static_cast<double>(infoStruct->NumberOfThreads);
    double dataSize = static_cast<double>(caller->data.size());
    unsigned int startSearchIndex = static_cast<unsigned int>(threadID*dataSize/numThreads); 
    unsigned int endSearchIndex = static_cast<unsigned int>((threadID+1)*dataSize/numThreads-1);
    unsigned int myMaxValueIndex = startSearchIndex;
    T myMaxValue = caller->data[myMaxValueIndex];
             //print the thread ID and it's search range
    caller->mutex.Lock();
    std::cout<<"Thread #"<<threadID<<", index search range: ["<<startSearchIndex<<", "<<endSearchIndex<<"]\n";
    caller->mutex.Unlock(); 

            //find local max
    for(unsigned int i=startSearchIndex+1; i<=endSearchIndex; i++)
    {
      if(caller->data[i]>myMaxValue)
      {
        myMaxValueIndex = i;
        myMaxValue = caller->data[myMaxValueIndex]; 
      }
    }
        //lock the mutex, check if local max greater than the common max
    caller->mutex.Lock();
    if(caller->data[caller->maxIndex] < myMaxValue)
      caller->maxIndex = myMaxValueIndex;
    caller->mutex.Unlock();
  } 
  return ITK_THREAD_RETURN_VALUE;
}
